@@ -32,6 +32,11 @@
 #include <structmember.h>
 #include <datetime.h>
 
+// Py_ssize_t is defined starting from Python 2.5
+#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
+typedef int Py_ssize_t;
+#endif
+
 // lame assumption, but for now I have no better idea
 #ifndef MS_WINDOWS
 #define HAVE_FREETDS
@@ -44,9 +49,14 @@
 #include <lmerr.h>
 #include <sqlfront.h>
 #include <sqldb.h>     // DB-LIB header file (should always be included)
+
+#ifndef SQLINT8
+#define SQLINT8 127    // from freetds/include/sybdb.h
+#endif
+
 #else
 #define MSDBLIB        // we need FreeTDS to provide MSSQL API,
-// not Sybase API. See README.freetds for details
+                       // not Sybase API. See README.freetds for details
 #include <sqlfront.h>
 #include <sqldb.h>
 
@@ -56,6 +66,7 @@
 #define SQLINT1        SYBINT1
 #define SQLINT2        SYBINT2
 #define SQLINT4        SYBINT4
+#define SQLINT8        SYBINT8
 #define SQLINTN        SYBINTN
 #define SQLFLT4        SYBREAL
 #define SQLFLT8        SYBFLT8
@@ -230,6 +241,7 @@ static PyObject *_mssql_select_db(_mssql_connection *self, PyObject *args) {
 	RETCODE rtc;
 	char *dbname;
 	char command[255];
+	PyObject *value, *search, *replace, *replaced = NULL;
 
 	if (PyErr_Occurred())  return NULL;
 	assert_connected(self);
@@ -240,10 +252,22 @@ static PyObject *_mssql_select_db(_mssql_connection *self, PyObject *args) {
 	/* this workaround is for db name truncation by ntwdblib.dll's
 	   version of dbuse(), thanks Luke Benstead <luke@riverhall.co.uk> */
 
-	if (dbname[0] == '[')
+	/* if name starts in '[' and ends in ']' then we pass it as is;
+	   otherwise we replace all occurrences of ']' with ']]' then add
+	   '[' at the beginning and ']' at the end. */
+
+	if ((dbname[0] == '[') && dbname[strlen(dbname)-1] == ']')
 		snprintf(command, 255, "USE %s", dbname);
-	else
+	else {
+		value = PyString_FromString(dbname);
+		search = PyString_FromString("]");
+		replace = PyString_FromString("]]");
+		replaced = PyObject_CallMethod(value, "replace", "OO", search, replace);
+		dbname = PyString_AsString(replaced);
+		Py_DECREF(value); Py_DECREF(search); Py_DECREF(replace);
 		snprintf(command, 255, "USE [%s]", dbname);
+		Py_DECREF(replaced);
+	}
 
 	Py_BEGIN_ALLOW_THREADS
 	rtc = dbcmd(self->dbproc, command);
@@ -348,6 +372,17 @@ static PyObject *_mssql_execute_scalar(_mssql_connection *self, PyObject *args)
 	rtc = dbnextrow(self->dbproc);
 	Py_END_ALLOW_THREADS
 
+	check_cancel_and_raise(rtc, self);
+
+	self->rows_affected = dbcount(self->dbproc); // get rows affected, if possible
+
+	// check if any rows (in case user passed e.g. DDL command here)
+	if (rtc == NO_MORE_ROWS) {
+		clr_metadata(self);
+		self->last_dbresults = 0;  // force next results, if any
+		Py_RETURN_NONE;
+	}
+
 	row = get_row(self, rtc); // get first row of data
 	if (row == NULL)  return NULL;
 
@@ -508,7 +543,8 @@ PyObject *_quote_or_flatten(PyObject *data) {
 
 	if (PyList_Check(data)) {
 		PyObject *str;
-		int i, len = PyList_GET_SIZE(data);
+		int i;
+		Py_ssize_t len = PyList_GET_SIZE(data);
 
 		str = PyString_FromString("");  // an empty string
 		if (str == NULL)  return NULL;
@@ -551,7 +587,8 @@ PyObject *_quote_or_flatten(PyObject *data) {
 
 	if (PyTuple_Check(data)) {
 		PyObject *str;
-		int i, len = PyTuple_GET_SIZE(data);
+		int i;
+		Py_ssize_t len = PyTuple_GET_SIZE(data);
 
 		str = PyString_FromString("");  // an empty string
 		if (str == NULL)  return NULL;
@@ -634,7 +671,7 @@ PyObject *_mssql_quote_data(PyObject * /*unused*/ self, PyObject *data) {
 	// first check if dict, and if so, quote values
 	if (PyDict_Check(data)) {
 		PyObject *dict, *k, *v;
-		int pos = 0;
+		Py_ssize_t pos = 0;
 		dict = PyDict_New();                       // new dictionary for results
 		if (dict == NULL)  return NULL;
 
@@ -655,7 +692,8 @@ PyObject *_mssql_quote_data(PyObject * /*unused*/ self, PyObject *data) {
 
 	if (PyTuple_Check(data)) {                     // input is a tuple
 		PyObject *res;
-		int i, len = PyTuple_GET_SIZE(data);
+		int i;
+		Py_ssize_t len = PyTuple_GET_SIZE(data);
 
 		res = PyTuple_New(len);
 		if (res == NULL)  return NULL;
@@ -817,7 +855,7 @@ static PyObject *_mssql_nextresult(_mssql_connection *self, PyObject *args) {
 		check_cancel_and_raise(rtc, self);
 	}
 
-	self->last_dbresults = 0;  // force call to dbresults()
+	self->last_dbresults = 0;  // force call to dbresults() inside get_result()
 
 	if (get_result(self) == NULL)  return NULL;
 
@@ -985,7 +1023,15 @@ PyObject *_mssql_identity_get(_mssql_connection *self, void *closure) {
 	rtc = dbnextrow(self->dbproc);
 	Py_END_ALLOW_THREADS
 
-	row = get_row(self, rtc); // get first row of data
+	check_cancel_and_raise(rtc, self);
+
+	if (rtc == NO_MORE_ROWS) {     // check it, just to be pendatically sure
+		clr_metadata(self);
+		self->last_dbresults = 0;  // force next results, if any
+		Py_RETURN_NONE;
+	}
+
+	row = get_row(self, rtc);      // get first row of data
 	if (row == NULL)  return NULL;
 
 	id = PyTuple_GetItem(row, 0);
@@ -1213,8 +1259,10 @@ password - password to authenticate with.\n\
 trusted  - use trusted connection (Windows Integrated Authentication)\n\
            instead of SQL user and password, user and password\n\
            are ignored. Only available on Windows.\n\
-charset  - character set name to set for the connection. [XXX BROKEN]\n\
+charset  - character set name to set for the connection.\n\
 database - database name to select initially as the current database.\n\
+max_conn - maximum number of simultaneous connections allowed; default\n\
+           is 25\n\
 ";
 
 static PyObject *_mssql_connect(_mssql_connection *self, PyObject *args,
@@ -1224,20 +1272,25 @@ static PyObject *_mssql_connect(_mssql_connection *self, PyObject *args,
 	char *server = NULL, *user = NULL, *password = NULL;
 	char *database = NULL, *charset = NULL;
 	char *p;
-	int trusted = 0;
+	int trusted = 0, max_conn = 25;
 	RETCODE rtc;
 	struct _mssql_connection_list_node *n;
 	PyObject *ologintimeout, *o;
 
 	static char *kwlist[] = { "server", "user", "password", "trusted",
-			"charset", "database", NULL };
+			"charset", "database", "max_conn", NULL };
 
-	if(!PyArg_ParseTupleAndKeywords(args, kwargs, "s|ssiss:connect", kwlist,
-			&server, &user, &password, &trusted, &charset, &database))
+	if(!PyArg_ParseTupleAndKeywords(args, kwargs, "s|zzizzi:connect", kwlist,
+			&server, &user, &password, &trusted, &charset, &database, &max_conn))
 		return NULL;
 
 	clr_err(NULL);
 
+#ifdef PYMSSQLDEBUG
+	fprintf(stderr, "_mssql.connect(server=\"%s\", user=\"%s\", password=\"%s\", trusted=\"%d\", charset=\"%s\", database=\"%s\", max_conn=\"%d\")\n",
+			server, user, password, trusted, charset, database, max_conn);
+#endif
+
 	// lame hack to improve portability, will break with IPv6
 	// FreeTDS doesn't accept ',' as port number separator
 #ifdef MS_WINDOWS
@@ -1256,10 +1309,14 @@ static PyObject *_mssql_connect(_mssql_connection *self, PyObject *args,
 	if (login == NULL)
 		raise_MssqlDriverException("Out of memory");
 
+	if (max_conn < 0)
+		raise_MssqlDriverException("max_conn value must be greater than 0.");
+
 	// these don't need to release GIL
 	DBSETLUSER(login, user);
 	DBSETLPWD(login, password);
 	DBSETLAPP(login, "pymssql");
+	dbsetmaxprocs(max_conn);
 #ifdef MS_WINDOWS
 	DBSETLVERSION(login, DBVER60);
 #else
@@ -1405,18 +1462,12 @@ static void _mssql_connection_dealloc(_mssql_connection *self) {
 
 	Py_XDECREF(self->column_names);
 	Py_XDECREF(self->column_types);
-	PyObject_Free((char *) self);
+	PyObject_Free((void *) self);
 }
 
 static PyObject *_mssql_connection_repr(_mssql_connection *self) {
-	char buf[100];
-
-	if(self->connected)
-		snprintf(buf, 100, "<Open mssql connection at %lx>", (long) self);
-	else
-		snprintf(buf, 100, "<Closed mssql connection at %lx>", (long) self);
-
-	return PyString_FromString(buf);
+	return PyString_FromFormat("<%s mssql connection at %p>",
+		self->connected ? "Open" : "Closed", self);
 }
 
 /* instance methods */
@@ -1458,7 +1509,7 @@ static PyMemberDef _mssql_connection_members[] = {
 	  "this value is only meaningful after reading all rows." },
 	{ "charset", T_STRING, _MSSQL_OFF(charset), READONLY,
 	  "Character set name that was passed to _mssql.connect()." },
-	{ "debug_queries", T_INT, _MSSQL_OFF(debug_queries), WRITE_RESTRICTED,
+	{ "debug_queries", T_INT, _MSSQL_OFF(debug_queries), RESTRICTED,
 	  "If set to True, all queries are printed to stderr after\n"
 	  "formatting and quoting, just before execution." },
 	{ NULL, 0, 0, 0, NULL },
@@ -1682,10 +1733,7 @@ init_mssql(void) {
 /* _mssql.MssqlRowIterator class methods *************************************/
 
 static PyObject *_mssql_row_iterator_repr(_mssql_row_iterator *self) {
-	char buf[100];
-
-	snprintf(buf, 100, "<_mssql.MssqlRowIterator at %lx>", (long) self);
-	return PyString_FromString(buf);
+	return PyString_FromFormat("<_mssql.MssqlRowIterator at %p>", self);
 }
 
 static PyObject *_mssql_row_iterator__iter__(PyObject *self) {
@@ -1803,18 +1851,21 @@ int rmv_lcl(char *s, char *buf, size_t buflen) {
 	}
 
 	*b = 0;
-	return (b - buf);  // return new len
+
+	// cast to int to make x64 happy, can do it because numbers are not so very long
+	return (int)(b - buf);  // return new len
 }
 
 /* This is a helper function, which does most work needed by any
    execute_*() function. It returns NULL on error, non-NULL on success.
-   Actually it returns Py_None, but there's no need to DECREF it. Also,
-   we can't use ints here because of assert_connected and other macros. */
+   Actually it returns Py_None, but there's no need to DECREF it. Also, we
+   can't use int as retcode here because assert_connected and other macros
+   return NULL on failure. */
 
 PyObject *format_and_run_query(_mssql_connection *self, PyObject *args) {
 	RETCODE rtc;
 	char *query;
-	PyObject *format = NULL, *params = NULL, *formatted, *tuple;
+	PyObject *format = NULL, *params = NULL, *formatted = NULL, *tuple;
 
 	if (PyErr_Occurred())  return NULL;
 	assert_connected(self);
@@ -1824,6 +1875,7 @@ PyObject *format_and_run_query(_mssql_connection *self, PyObject *args) {
 		return NULL;
 
 	// format the string then convert it to char *
+	// Warning: don't put params != Py_None here as None is a valid value!
 	if (params != NULL) {
 		tuple = PyTuple_New(2);
 		if (tuple == NULL)  return NULL;
@@ -1834,8 +1886,7 @@ PyObject *format_and_run_query(_mssql_connection *self, PyObject *args) {
 		formatted = _mssql_format_sql_command(NULL, tuple);
 		Py_DECREF(tuple);
 		if (formatted == NULL)  return NULL;
-		query = PyString_AsString(formatted);
-		Py_DECREF(formatted);
+		query = PyString_AsString(formatted); // remember to DECREF later
 	} else {
 		query = PyString_AsString(format);
 	}
@@ -1855,6 +1906,10 @@ PyObject *format_and_run_query(_mssql_connection *self, PyObject *args) {
 	rtc = dbsqlexec(self->dbproc);
 	Py_END_ALLOW_THREADS
 
+	// remember to DECREF this as soon as 'query' is no longer needed
+	// and before any possible return statement
+	Py_XDECREF(formatted);
+
 	check_cancel_and_raise(rtc, self);
 
 	return Py_None;  // borrowed, but it doesn't matter, it's just a flag
@@ -1864,7 +1919,9 @@ PyObject *format_and_run_query(_mssql_connection *self, PyObject *args) {
    setting column_names and column_types properties. If the result appears
    to be already read, function does nothing. It returns NULL on error and
    non-NULL on success. Actually it returns Py_None, but there's no need
-   to DECREF it. */
+   to DECREF it. This function skips over all result sets that have no
+   columns (for example the query: "DECLARE @a INT; SELECT 1" returns
+   two result sets: one with zero columns, followed by one with one column. */
 
 PyObject *get_result(_mssql_connection *conn) {
 	int col;
@@ -1875,9 +1932,14 @@ PyObject *get_result(_mssql_connection *conn) {
 
 	clr_metadata(conn);
 
-	Py_BEGIN_ALLOW_THREADS
-	conn->last_dbresults = dbresults(conn->dbproc);
-	Py_END_ALLOW_THREADS
+	// find a result set that has at least one column
+	conn->last_dbresults = SUCCEED;
+	while (conn->last_dbresults == SUCCEED &&
+			(conn->num_columns = dbnumcols(conn->dbproc)) <= 0) {
+		Py_BEGIN_ALLOW_THREADS
+		conn->last_dbresults = dbresults(conn->dbproc);
+		Py_END_ALLOW_THREADS
+	}
 
 	check_cancel_and_raise(conn->last_dbresults, conn);
 
@@ -1902,7 +1964,8 @@ PyObject *get_result(_mssql_connection *conn) {
 		coltype = dbcoltype(conn->dbproc, col);
 
 		switch (coltype) {
-			case SQLBIT: case SQLINT1: case SQLINT2: case SQLINT4: case SQLINTN:
+			case SQLBIT: case SQLINT1: case SQLINT2: case SQLINT4:
+			case SQLINT8: case SQLINTN:
 			case SQLFLT4: case SQLFLT8: case SQLFLTN:
 				apicoltype = TYPE_NUMBER;
 				break;
@@ -1963,6 +2026,7 @@ PyObject *get_row(_mssql_connection *conn, int rowinfo) {
 	int col, coltype, len;
 	LPBYTE data;                // column data pointer
 	long intdata;
+	PY_LONG_LONG longdata;
 	double ddata;
 	PyObject *record;
 	DBDATEREC di;
@@ -1986,10 +2050,14 @@ PyObject *get_row(_mssql_connection *conn, int rowinfo) {
 
 	for (col = 1; col <= conn->num_columns; col++) {     // do for all columns
 
-#ifdef PYMSSQLDEBUG
-		fprintf(stderr, "Processing row %d, column %d\n", DEBUGRowNumber, col);
-#endif
+//#ifdef PYMSSQLDEBUG
+//		fprintf(stderr, "Processing row %d, column %d\n", DEBUGRowNumber, col);
+//#endif
 		Py_BEGIN_ALLOW_THREADS
+
+		// get pointer to column's data
+		// rowinfo == FAIL and rowinfo == NO_MORE_ROWS are already handled in caller
+
 		data = GET_DATA(dbproc, rowinfo, col);    // get pointer to column's data
 		coltype = GET_TYPE(dbproc, rowinfo, col);
 		Py_END_ALLOW_THREADS
@@ -2000,6 +2068,13 @@ PyObject *get_row(_mssql_connection *conn, int rowinfo) {
 			continue;
 		}
 
+#ifdef PYMSSQLDEBUG
+		fprintf(stderr, "Processing row %d, column %d. Got data=%x, coltype=%d, len=%d\n",
+				DEBUGRowNumber, col, data, coltype,
+				GET_LEN(dbproc, rowinfo, col));
+
+#endif
+
 		switch (coltype) {                        // else we have data
 			case SQLBIT:
 				intdata = (int) *(DBBIT *) data;
@@ -2021,6 +2096,11 @@ PyObject *get_row(_mssql_connection *conn, int rowinfo) {
 				PyTuple_SET_ITEM(record, col-1, PyInt_FromLong((long) intdata));
 				break;
 
+			case SQLINT8:
+				longdata = *(PY_LONG_LONG *) data;
+				PyTuple_SET_ITEM(record, col-1, PyLong_FromLongLong((PY_LONG_LONG) longdata));
+				break;
+
 			case SQLFLT4:
 				ddata = *(DBFLT4 *)data;
 				PyTuple_SET_ITEM(record, col-1, PyFloat_FromDouble(ddata));
@@ -2088,11 +2168,9 @@ PyObject *get_row(_mssql_connection *conn, int rowinfo) {
 				break;
 
 			case SQLVARCHAR: case SQLCHAR: case SQLTEXT:
-				// on *nix with FreeTDS it's FreeTDS who converts data
-				// into UTF-8. On Windows we must convert them ourselves, using
-				// encoding provided by the user during connect()
-#ifdef MS_WINDOWS
-				if (*conn->charset)  // if charset given then decode
+				// Convert the strings from the encoding provided by the user
+				// If there is none, just return the strings as is.
+				if (*conn->charset)
 					PyTuple_SET_ITEM(record,col-1,PyUnicode_Decode((char *) data,
 							GET_LEN(dbproc,rowinfo,col), conn->charset, NULL));
 				else
@@ -2100,10 +2178,6 @@ PyObject *get_row(_mssql_connection *conn, int rowinfo) {
 					PyTuple_SET_ITEM(record, col-1, PyString_FromStringAndSize(
 							(const char *)data, GET_LEN(dbproc, rowinfo, col)));
 
-#else
-				PyTuple_SET_ITEM(record,col-1,PyUnicode_Decode((char *) data,
-						GET_LEN(dbproc,rowinfo,col), "utf8", NULL));
-#endif
 				break;
 
 			//case SQLBINARY: case SQLVARBINARY: case SQLIMAGE:
@@ -2143,6 +2217,8 @@ PyObject *fetch_next_row_dict(_mssql_connection *conn, int raise) {
 	rtc = dbnextrow(conn->dbproc);
 	Py_END_ALLOW_THREADS
 
+	check_cancel_and_raise(rtc, conn);
+
 	if (rtc == NO_MORE_ROWS) {
 		clr_metadata(conn);
 		// 'rows affected' is nonzero only after all records are read
